"""
Draft Sport Python
Human Module
author: hugh@blinkybeach.com
"""
from nozomi import Immutable, Configuration, RequestCredentials
from nozomi import Decodable
from typing import Optional, Type, TypeVar, Any
from nozomi import URLParameter, URLParameters, HTTPMethod, ApiRequest
from draft_sport.communication.method import CommunicationMethod

T = TypeVar('T', bound='Human')


class Human(Decodable):

    _PATH = '/human'

    def __init__(
        self,
        public_id: str,
        email: CommunicationMethod
    ) -> None:

        self._email = email
        self._public_id = public_id

        return

    display_name: str = Immutable(lambda s: s._email.body.split('@')[0])
    public_id: str = Immutable(lambda s: s._public_id)
    email: CommunicationMethod = Immutable(lambda s: s._email)

    @classmethod
    def retrieve(
        cls: Type[T],
        public_id: str,
        credentials: RequestCredentials,
        configuration: Configuration
    ) -> Optional[T]:
        """
        Optionally return a Human with the given public ID, if it exists
        """

        assert isinstance(public_id, str)

        target = URLParameter('public_id', public_id)
        parameters = URLParameters([target])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            configuration=configuration,
            data=None,
            url_parameters=parameters,
            credentials=credentials
        )

        return cls.optionally_decode(request.response_data)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            public_id=data['public_id'],
            email=CommunicationMethod.decode(data['email'])
        )



"""
Draft Sport Python
Point Score Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import TypeVar, Type, Any

T = TypeVar('T', bound='Score')


class Score(Decodable):
    """An integer score with respect to some fantasy metric"""

    def __init__(
        self,
        fantasy_metric_name: str,
        score: int
    ) -> None:

        self._fantasy_metric_name = fantasy_metric_name
        self._score = score

        return

    value = Immutable(lambda s: s._score)
    fantasy_metric_name = Immutable(lambda s: s._fantasy_metric_name)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            fantasy_metric_name=data['fantasy_metric_name'],
            score=data['score']
        )



"""
Draft Sport Python Library
Offer Module
author: hugh@blinkybeach.com
"""
from typing import TypeVar, Type, Any, List, Optional, Dict
from decimal import Decimal
from nozomi import Decodable
from draft_sport.store.price import Price
from draft_sport.store.product import Product
from nozomi import Immutable, Configuration, RequestCredentials
from nozomi import URLParameter, URLParameters, HTTPMethod, ApiRequest
from draft_sport.store.frequency import Frequency

T = TypeVar('T', bound='Offer')


class Offer(Decodable):

    _PATH = '/offer'

    def __init__(
        self,
        public_id: str,
        name: str,
        prices: List[Price],
        products: List[Product],
        frequency: Frequency
    ) -> None:

        self._public_id = public_id
        self._name = name
        self._prices = prices
        self._products = products
        self._frequency = frequency

        return

    public_id: str = Immutable(lambda s: s._public_id)
    name: str = Immutable(lambda s: s._name)
    prices: List[Price] = Immutable(lambda s: s._prices)
    products: List[Product] = Immutable(lambda s: s._products)
    features: List[str] = Immutable(lambda s: s._flatten_features())
    frequency: Frequency = Immutable(lambda s: s._frequency)
    price_map: Dict[str, Decimal] = Immutable(
        lambda s: {p.denomination.iso_4217: str(p.magnitude) for p in s._prices}
    )

    def _flatten_features(self) -> List[str]:
        flat_features: List[str] = list()
        for product in self._products:
            for feature in product.features:
                flat_features.append(feature)
        return flat_features

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            public_id=data['public_id'],
            name=data['name'],
            prices=Price.decode_many(data['prices']),
            products=Product.decode_many(data['products']),
            frequency=Frequency(data['frequency'])
        )

    @classmethod
    def retrieve(
        cls: Type[T],
        public_id: str,
        credentials: RequestCredentials,
        configuration: Configuration
    ) -> Optional[T]:

        parameters = URLParameters([URLParameter('offer', public_id)])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            configuration=configuration,
            data=None,
            url_parameters=parameters,
            credentials=credentials
        )

        return cls.optionally_decode(request.response_data)



"""
Draft Sport Python
Player Points Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from draft_sport.fantasy.scores.player.round import Round
from draft_sport.fantasy.scores.player.score import Score
from typing import TypeVar, Type, Any, List, Dict, Set

T = TypeVar('T', bound='Points')


class Points(Decodable):

    def __init__(
        self,
        average_points: int,
        total_points: int,
        points_last_round: int,
        points_per_minute_played: int,
        rounds: List[Round]
    ) -> None:

        self._average_points = average_points
        self._total_points = total_points
        self._points_last_round = points_last_round
        self._points_per_minute_played = points_per_minute_played
        self._rounds = sorted(rounds, key=lambda r: r.round_sequence)

        self._cached_flattened_rounds: Optional[Dict[str, List[int]]] = None

        return

    average_points = Immutable(lambda s: s._average_points)
    total_points = Immutable(lambda s: s._total_points)
    points_last_round = Immutable(lambda s: s._points_last_round)
    points_per_minute_played = Immutable(lambda s: s._points_per_minute_played)
    rounds = Immutable(lambda s: s._rounds)

    normalised_rounds = Immutable(lambda s: s._compute_normalised_rounds())
    flattened_rounds = Immutable(lambda s: s._retrieved_flattened_rounds())

    _all_metric_keys = Immutable(lambda s: s._compute_all_metric_keys())

    def _compute_all_metric_keys(self) -> Set[str]:
        keys: Set[str] = set()
        for rou in self._rounds:
            for score in rou.scores:
                keys.add(score.fantasy_metric_name)
        return sorted(keys)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            points_per_minute_played=data['points_per_minute_played'],
            average_points=data['average_points'],
            total_points=data['total_points'],
            points_last_round=data['points_last_round'],
            rounds=Round.decode_many(data['rounds'])
        )

    def _compute_normalised_rounds(self) -> List[Round]:

        all_keys = self._all_metric_keys

        def _scores_for_round(rou: Round) -> List[Score]:

            scores: List[Score] = list()

            for key in all_keys:

                found = False

                for score in rou.scores:
                    if score.fantasy_metric_name == key:
                        scores.append(score)
                        found = True
                        break
                    
                if found is True:
                    break

                scores.append(Score(
                    fantasy_metric_name=key,
                    score=0
                ))

            return sorted(scores, key=lambda s: s.fantasy_metric_name)

        rounds: List[Round] = list()

        for rou in self._rounds:

            rounds.append(Round(
                sequence=rou.round_sequence,
                scores=_scores_for_round(rou)
            ))

        return sorted(rounds, key=lambda r: r.round_sequence)

    def _retrieved_flattened_rounds(self) -> Dict[str, List[int]]:
        if self._cached_flattened_rounds is None:
            self._cached_flattened_rounds = self._compute_flattened_rounds()
        return self._cached_flattened_rounds

    def _compute_flattened_rounds(self) -> Dict[str, List[int]]:

        flat_rounds: Dict[str, List[int]] = dict()

        for key in self._all_metric_keys:

            metric_stats: List[int] = list()

            for rou in self._rounds:

                found = False
                
                for score in rou.scores:
                    if score.fantasy_metric_name == key:
                        metric_stats.append(score.value)
                        found = True
                        break
                    continue

                if found is True:
                    continue

                metric_stats.append(0)
                continue

            flat_rounds[key] = metric_stats

            continue

        return flat_rounds



"""
Draft Sport
Name Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import TypeVar, Type, Any

T = TypeVar('T', bound='Name')


class Name(Decodable):

    def __init__(
        self,
        first_name: str,
        other_name: str,
        last_name: str
    ) -> None:

        self._first_name = first_name
        self._other_name = other_name
        self._last_name = last_name

        return

    first = Immutable(lambda s: s._first_name)
    other = Immutable(lambda s: s._other_name)
    last = Immutable(lambda s: s._last_name)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            first_name=data['first_name'],
            other_name=data['other_name'],
            last_name=data['last_name']
        )



"""
Draft Sport Python Library
Price Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import Type, TypeVar, Any
from decimal import Decimal
from draft_sport.store.denomination import Denomination

T = TypeVar('T', bound='Price')


class Price(Decodable):

    def __init__(
        self,
        magnitude: Decimal,
        denomination: Denomination
    ) -> None:

        self._magnitude = magnitude
        self._denomination = denomination

        return

    magnitude = Immutable(lambda s: s._magnitude)
    denomination = Immutable(lambda s: s._denomination)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            magnitude=Decimal(data['magnitude']),
            denomination=Denomination.decode(data['denomination'])
        )



"""
Draft Sport Python Library
Challenge Module
author: hugh@blinkybeach.com
"""
from nozomi import Immutable, Configuration, RequestCredentials
from nozomi import Decodable, NozomiTime
from typing import TypeVar, Type, Any, Optional
from nozomi import URLParameter, URLParameters, HTTPMethod, ApiRequest

T = TypeVar('T', bound='Challenge')


class Challenge(Decodable):

    _PATH = '/communication-method/challenge'

    def __init__(
        self,
        expiration: NozomiTime,
        completed: Optional[NozomiTime],
        is_expired: bool,
        method_id: str
    ) -> None:

        self._expiration = expiration
        self._is_expired = is_expired
        self._completed = completed
        self._method_id = method_id

        return

    is_completed: bool = Immutable(lambda s: s._completed is not None)
    is_expired: bool = Immutable(lambda s: s._is_expired)
    method_id: str = Immutable(lambda s: s._method_id)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            expiration=NozomiTime.decode(data['expiration']),
            completed=NozomiTime.optionally_decode(data['completed']),
            is_expired=data['is_expired'],
            method_id=data['method_id']
        )

    @classmethod
    def retrieve(
        cls: Type[T],
        code: str,
        credentials: RequestCredentials,
        configuration: Configuration
    ) -> Optional[T]:

        parameters = URLParameters([
            URLParameter('code', code)
        ])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            configuration=configuration,
            data=None,
            url_parameters=parameters,
            credentials=credentials
        )

        return cls.optionally_decode(request.response_data)



"""
Draft Sport Python
Perspective Module
author: hugh@blinkybeach.com
"""
from nozomi import Perspective as NozomiPerspective
from nozomi import Immutable
from typing import TypeVar, Type

T = TypeVar('T', bound='Perspective')


class Perspective(NozomiPerspective):

    def __init__(
        self,
        perspective_id: int,
        perspective_name: str
    ) -> None:

        self._perspective_id = perspective_id
        self._perspective_name = perspective_name

        return

    perspective_id = Immutable(lambda s: s._perspective_id)
    value = Immutable(lambda s: s._perspective_id)

    @classmethod
    def with_id(cls: Type[T], perspective_id: int) -> T:
        if perspective_id == Constants.ADMINISTRATOR.perspective_id:
            return Constants.ADMINISTRATOR
        if perspective_id == Constants.CUSTOMER.perspective_id:
            return Constants.CUSTOMER
        raise RuntimeError('Unknown Perspective ID')

    def __hash__(self) -> int:
        return hash(self._perspective_id)


class Constants:

    ADMINISTRATOR = Perspective(2, 'Administrator')
    CUSTOMER = Perspective(1, 'Customer')



"""
Draft Sport Python Library
Product Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import TypeVar, Type, Any, List

T = TypeVar('T', bound='Product')


class Product(Decodable):

    def __init__(
        self,
        public_id: str,
        name: str,
        features: List[str]
    ) -> None:

        self._public_id = public_id
        self._name = name
        self._features = features

        return

    features = Immutable(lambda s: s._features)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            public_id=data['public_id'],
            name=data['name'],
            features=data['features']
        )



"""
Draft Sport Python
Fantasy Player Profile Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import TypeVar, Type, Any

T = TypeVar('T', bound='Profile')


class Profile(Decodable):

    def __init__(
        self,
        first_name: str,
        last_name: str,
        position_name: str,
        public_id: str,
        team_name: str
    ) -> None:

        self._first_name = first_name
        self._last_name = last_name
        self._position_name = position_name
        self._public_id = public_id
        self._team_name = team_name

        return

    first_name = Immutable(lambda s: s._first_name)
    last_name = Immutable(lambda s: s._last_name)
    public_id = Immutable(lambda s: s._public_id)
    position_name = Immutable(lambda s: s._position_name)
    team_name = Immutable(lambda s: s._team_name)

    full_name = Immutable(lambda s: s._first_name + ' ' + s._last_name)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            first_name=data['first_name'],
            last_name=data['last_name'],
            position_name=data['position_name'],
            public_id=data['public_id'],
            team_name=data['team_name']
        )



"""
Draft Sport
API Key Module
author: hugh@blinkybeach.com
"""
from draft_sport.errors.type import DSTypeError


class ApiKey:

    def __init__(self, key: str) -> None:

        DSTypeError.assert_type(key, 'key', str)

        self._key = key

        return



from draft_sport.leagues.league import League
from draft_sport.leagues.pick import Player
from draft_sport.leagues.team import Team
from draft_sport.leagues.invitation import Invitation

from draft_sport.fantasy.scores.player.player import Player
from draft_sport.fantasy.scores.player.order_by import OrderBy
from draft_sport.fantasy.scores.player.points import Points
from draft_sport.fantasy.scores.player.round import Round
from draft_sport.fantasy.scores.player.score import Score
from draft_sport.fantasy.player.profile import Profile

from draft_sport.data.order import Order

from draft_sport.security.session import Session
from draft_sport.security.perspective import Perspective

from draft_sport.humans.human import Human

from draft_sport.communication.challenge import Challenge
from draft_sport.communication.method import CommunicationMethod

from draft_sport.store.denomination import Denomination
from draft_sport.store.offer import Offer
from draft_sport.store.product import Product
from draft_sport.store.price import Price



"""
Draft Sport
Configuration Module
author: hugh@blinkybeach.com
"""


class Configuration:
    """
    Abstract protocol defining an interfaces for classes that may
    serve as a configuration store for Draft Sport Python.
    """

    api_endpoint: str = NotImplemented



"""
Draft Sport Python
Communication Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import TypeVar, Type, Any

T = TypeVar('T', bound='CommunicationMethod')


class CommunicationMethod(Decodable):

    def __init__(
        self,
        public_id: str,
        mode: int,
        body: str,
        confirmed: bool
    ) -> None:

        self._public_id = public_id
        self._mode = mode
        self._body = body
        self._confirmed = confirmed

        return

    body = Immutable(lambda s: s._body)
    public_id = Immutable(lambda s: s._public_id)
    confirmed = Immutable(lambda s: s._confirmed)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            public_id=data['public_id'],
            mode=data['mode'],
            body=data['body'],
            confirmed=data['confirmed']
        )



"""
Draft Sport Python Library
Denomination Module
author: hugh@blinkybeach.com
"""
from nozomi import Decodable, Immutable
from typing import Type, TypeVar, Any

T = TypeVar('T', bound='Denomination')


class Denomination(Decodable):

    def __init__(
        self,
        name: str,
        iso_4217: str,
        symbol: str
    ) -> None:

        self._name = name
        self._iso_4217 = iso_4217
        self._symbol = symbol

        return

    name: str = Immutable(lambda s: s._name)
    iso_4217: str = Immutable(lambda s: s._iso_4217)
    symbol: str = Immutable(lambda s: s._symbol)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            name=data['name'],
            iso_4217=data['iso_4217'],
            symbol=data['symbol']
        )



"""
Draft Sport
Team Module
author: hugh@blinkybeach.com
"""
from typing import List, Any, TypeVar, Type, Optional
from nozomi import Immutable, Decodable, RequestCredentials, ApiRequest
from nozomi import Configuration, URLParameter, URLParameters, HTTPMethod
from draft_sport.leagues.pick import Pick

T = TypeVar('T', bound='Team')


class Team(Decodable):

    _PATH = '/league/team'

    def __init__(
        self,
        league_id: str,
        picks: List[Pick],
        manager_id: str,
        manager_display_name: str,
        name: Optional[str],
        total_points: int
    ) -> None:

        self._league_id = league_id
        self._picks = picks
        self._manager_id = manager_id
        self._manager_display_name = manager_display_name
        self._name = name
        self._total_points = total_points

        return

    league_id = Immutable(lambda s: s._league_id)
    picks = Immutable(lambda s: s._picks)
    manager_id = Immutable(lambda s: s._manager_id)
    manager_display_name = Immutable(lambda s: s._manager_display_name)
    name = Immutable(lambda s: s._name)
    total_points = Immutable(lambda s: s._total_points)

    @classmethod
    def decode(cls: Type[T], data: Any) -> T:
        return cls(
            league_id=data['league_id'],
            picks=Pick.decode_many(data['picks']),
            manager_id=data['manager_id'],
            manager_display_name=data['manager_display_name'],
            name=data['name'],
            total_points=data['total_points']
        )

    @classmethod
    def retrieve(
        cls: Type[T],
        league_id: str,
        manager_id: str,
        credentials: RequestCredentials,
        configuration: Configuration
    ) -> Optional[T]:
        """
        Return a Team in the supplied League, managed by the supplied Manager,
        if it exists.
        """

        parameters = URLParameters([
            URLParameter('league', league_id),
            URLParameter('manager', manager_id)
        ])

        request = ApiRequest(
            path=cls._PATH,
            method=HTTPMethod.GET,
            configuration=configuration,
            data=None,
            url_parameters=parameters,
            credentials=credentials
        )

        return cls.optionally_decode(request.response_data)
